from collections import defaultdict
import streamlit as st
import pandas as pd
import difflib
import numpy as np
import jiwer
from evaluate import load
import argparse


st.set_page_config(
    layout="wide",
)


# Function to compute text diff
def compute_diff(ground_truth, prediction):
    diff = difflib.ndiff(ground_truth, prediction)
    diff_str = "".join(diff)
    return diff_str


# Function to compute GitHub-style visual diff
def compute_visual_diff(ground_truth, prediction):
    diff = difflib.ndiff(ground_truth, prediction)
    diff_html = []
    for line in diff:
        if line.startswith("+"):
            diff_html.append(f'<span style="color:green">{line[1:]}</span>')
        elif line.startswith("-"):
            diff_html.append(f'<span style="color:red">{line[1:]}</span>')
        elif line.startswith(" "):
            diff_html.append(line[1:])
    return "".join(diff_html)


# Set up command line argument parsing
def parse_args():
    parser = argparse.ArgumentParser(description="Streamlit app for OCR diagnostics")
    parser.add_argument(
        "--data-path",
        type=str,
        required=True,
        help="Path to the CSV file containing predictions.",
    )
    return parser.parse_args()


# Load CSV data and adjust image paths
@st.cache_data
def load_data(data_path):
    data = pd.read_csv(data_path)
    return data


args = parse_args()

data = load_data(args.data_path)


# Function to find and group duplicate gt and pred pairs
def find_duplicates(data):
    duplicates = data[data.duplicated(["gt", "pred"], keep=False)]
    grouped = (
        duplicates.groupby(["gt", "pred"]).apply(lambda x: list(x.index)).reset_index()
    )
    grouped.columns = ["gt", "pred", "image_indices"]
    return grouped


# Set up initial state
if "index" not in st.session_state:
    st.session_state.index = 0
if "group_indices" not in st.session_state:
    st.session_state.group_indices = []


def next_image():
    if st.session_state.group_indices:
        current_pos = st.session_state.group_indices.index(st.session_state.index)
        if current_pos < len(st.session_state.group_indices) - 1:
            st.session_state.index = st.session_state.group_indices[current_pos + 1]
    else:
        if st.session_state.index < len(data) - 1:
            st.session_state.index += 1


def prev_image():
    if st.session_state.group_indices:
        current_pos = st.session_state.group_indices.index(st.session_state.index)
        if current_pos > 0:
            st.session_state.index = st.session_state.group_indices[current_pos - 1]
    else:
        if st.session_state.index > 0:
            st.session_state.index -= 1


def jump_to_image(index):
    if 0 <= index < len(data):
        st.session_state.index = index
        st.session_state.group_indices = []


def select_group(indices):
    if indices:
        st.session_state.index = indices[0]
        st.session_state.group_indices = indices


st.sidebar.title(f"OCR Diagnostic App")

cer_evaluate = load("cer")
cer_score = cer_evaluate.compute(predictions=data["pred"], references=data["gt"])
st.sidebar.write(f"Avg CER: {cer_score:.5f}")

# Jump to specific image number
jump_to = st.sidebar.number_input(
    "Jump to image number:", min_value=1, max_value=len(data), step=1, value=1
)
if st.sidebar.button("Go"):
    jump_to_image(jump_to - 1)

# Display the current image and information
current_image = data.iloc[st.session_state.index]

# Compute diff between ground truth and prediction
diff_text = compute_diff(current_image["gt"], current_image["pred"])

visual_diff_html_1 = compute_visual_diff(
    " ".join((["_" if char == " " else char for char in list(current_image["gt"])])),
    " ".join((["_" if char == " " else char for char in list(current_image["pred"])])),
)

st.write(f"Image Path: {current_image['img_path']}")
st.image(current_image["img_path"], caption="Current Image")
st.write(f"CER: {current_image['CER']}")

# Set font size
font_size = 40

# Collapsible section for Ground Truth and Prediction
with st.expander("Ground Truth and Prediction"):
    st.markdown(
        f"<div style='overflow-x: auto; white-space: nowrap; font-size: {font_size}px; line-height: 1.5;'>{''.join((['_' if char == ' ' else char for char in list(current_image['gt'])]))}<br>{''.join((['_' if char == ' ' else char for char in list(current_image['pred'])]))}</div>",
        unsafe_allow_html=True,
    )

# Collapsible section for Ground Truth List, Predicted List, and Diff
with st.expander("Ground Truth List, Predicted List, and Diff"):
    st.markdown(
        f"<div style='overflow-x: auto; white-space: nowrap; font-size: {font_size}px; line-height: 1.5;'>{' '.join((['_' if char == ' ' else char for char in list(current_image['gt'])]))}<br>{' '.join((['_' if char == ' ' else char for char in list(current_image['pred'])]))}<br>{visual_diff_html_1}</div>",
        unsafe_allow_html=True,
    )


def run_substitutions():
    datacopy = data
    sub_list = []
    for i in range(len(datacopy)):
        current_image = datacopy.iloc[i]
        reference = current_image["gt"]
        hypothesis = current_image["pred"]
        out = jiwer.process_characters(reference, hypothesis)
        for A in out.alignments[0]:
            if A.type == "substitute":
                for ref_idx, hyp_idx in zip(
                    range(A.ref_start_idx, A.ref_end_idx),
                    range(A.hyp_start_idx, A.hyp_end_idx),
                ):
                    ref_unicode = ord(reference[ref_idx])
                    hyp_unicode = ord(hypothesis[hyp_idx])
                    sub_list.append([ref_unicode, hyp_unicode, i])
    return sub_list


def run_insertions():
    datacopy = data
    ins_list = []
    for i in range(len(datacopy)):
        current_image = datacopy.iloc[i]
        reference = current_image["gt"]
        hypothesis = current_image["pred"]
        out = jiwer.process_characters(reference, hypothesis)
        for A in out.alignments[0]:
            if A.type == "insert":
                for hyp_idx in range(A.hyp_start_idx, A.hyp_end_idx):
                    hyp_unicode = ord(hypothesis[hyp_idx])
                    ins_list.append([hyp_unicode, i])
    return ins_list


def run_deletions():
    datacopy = data
    del_list = []
    for i in range(len(datacopy)):
        current_image = datacopy.iloc[i]
        reference = current_image["gt"]
        hypothesis = current_image["pred"]
        out = jiwer.process_characters(reference, hypothesis)
        for A in out.alignments[0]:
            if A.type == "delete":
                for ref_idx in range(A.ref_start_idx, A.ref_end_idx):
                    ref_unicode = ord(reference[ref_idx])
                    del_list.append([ref_unicode, i])
    return del_list


# Convert to a NumPy array
data_array = run_substitutions()
data_array = np.array(data_array)

grouped = defaultdict(list)
for triplet in data_array:
    pair = (triplet[0], triplet[1])
    index = triplet[2]
    grouped[pair].append(index)

filtered_groups = {pair: indexes for pair, indexes in grouped.items()}
sorted_groups = dict(
    sorted(filtered_groups.items(), key=lambda item: len(item[1]), reverse=True)
)

total_images = 0
for pair, indexes in sorted_groups.items():
    total_images += len(indexes)

# print("Subs")
# for pair, indexes in sorted_groups.items():
#     print("SPACE" if hex(pair[0]) == hex(0x20) else chr(pair[0]), "SPACE" if hex(pair[1]) == hex(0x20) else chr(pair[1]), len(indexes), len(indexes) / total_images, pair[0], pair[1], hex(pair[0]), hex(pair[1]), abs(pair[0] - pair[1]))

with st.sidebar.expander("Substitutions"):
    # Display buttons for each filtered group with character pair
    indexes = [item for lst in list(sorted_groups.values())[:2] for item in lst]

    for pair, indexes in sorted_groups.items():
        character_pair = " ".join(
            chr(num) for num in pair
        )  # Convert Unicode to characters
        button_label = f"Unicode {pair[0]} {pair[1]} ({character_pair[::-1]}): {len(indexes)} Count"
        select = list(set(indexes))
        select.sort()
        if st.button(button_label):
            select_group(select)


# Group by the first element (Unicode decimal)
grouped = defaultdict(list)
for pair in run_insertions():
    unicode_value = pair[0]
    index = pair[1]
    grouped[unicode_value].append(index)

# Filter and sort groups
filtered_groups = {unicode_value: indexes for unicode_value, indexes in grouped.items()}
sorted_groups = dict(
    sorted(filtered_groups.items(), key=lambda item: len(item[1]), reverse=True)
)

total_images = 0
for pair, indexes in sorted_groups.items():
    total_images += len(indexes)

# print("Insertions")
# for pair, indexes in sorted_groups.items():
#     print("SPACE" if hex(pair) == hex(0x20) else chr(pair), len(indexes), len(indexes) / total_images, pair, hex(pair))


with st.sidebar.expander("Insertions"):
    indexes = [item for lst in list(sorted_groups.values())[:1] for item in lst]
    # Display buttons for each filtered group with character
    for unicode_value, indexes in sorted_groups.items():
        character = chr(unicode_value)  # Convert Unicode to character
        button_label = f"Unicode {unicode_value} ({character}): {len(indexes)} Count"
        select = list(set(indexes))
        select.sort()
        if st.button(button_label):
            select_group(select)

# Group by the first element (Unicode decimal)
grouped = defaultdict(list)
for pair in run_deletions():
    unicode_value = pair[0]
    index = pair[1]
    grouped[unicode_value].append(index)

# Filter and sort groups
filtered_groups = {unicode_value: indexes for unicode_value, indexes in grouped.items()}
sorted_groups = dict(
    sorted(filtered_groups.items(), key=lambda item: len(item[1]), reverse=True)
)


total_images = 0
for pair, indexes in sorted_groups.items():
    total_images += len(indexes)

# print("Deletions")
# for pair, indexes in sorted_groups.items():
#     print("SPACE" if hex(pair) == hex(0x20) else chr(pair), len(indexes), len(indexes) / total_images, pair, hex(pair))

with st.sidebar.expander("Deletions"):
    indexes = [item for lst in list(sorted_groups.values())[:1] for item in lst]
    # Display buttons for each filtered group with character
    for unicode_value, indexes in sorted_groups.items():
        character = chr(unicode_value)  # Convert Unicode to character
        button_label = f"Unicode {unicode_value} ({character}): {len(indexes)} Count"
        select = list(set(indexes))
        select.sort()
        if st.button(button_label):
            select_group(select)

# Navigation buttons
col1, col2, col3 = st.columns([1, 1, 1])
with col1:
    st.button("Previous", on_click=prev_image)
with col3:
    st.button("Next", on_click=next_image)

# Display the index
st.write(f"Image {st.session_state.index + 1} / {len(data)}")


reference = current_image["gt"]
hypothesis = current_image["pred"]
out = jiwer.process_characters(reference, hypothesis)

# Create three columns
col1, col2, col3 = st.columns(3)

# Substitutions
with col1:
    st.write("Substitutions")
    for A in out.alignments[0]:
        if A.type == "substitute":
            a = (
                "space"
                if hypothesis[A.hyp_start_idx] == " "
                else hypothesis[A.hyp_start_idx]
            )
            b = (
                "space"
                if reference[A.ref_start_idx] == " "
                else reference[A.ref_start_idx]
            )
            st.markdown(f"{a} {b}")

# Insertions
with col2:
    st.write("Insertions")
    for A in out.alignments[0]:
        if A.type == "insert":
            a = (
                "space"
                if hypothesis[A.hyp_start_idx] == " "
                else hypothesis[A.hyp_start_idx]
            )
            st.write(f"{a}")

# Deletions
with col3:
    st.write("Deletions")
    for A in out.alignments[0]:
        if A.type == "delete":
            a = (
                "space"
                if reference[A.ref_start_idx] == " "
                else reference[A.ref_start_idx]
            )
            st.write(f"{a}")
